import gc

import numpy as np
import h5py
from bistiming import IterTimer, SimpleTimer


def fill_concat_features(feature_list, global_feature_h5f, concat_feature_h5f,
                         buffer_size):
    feature_shapes = []
    for feature_name in feature_list:
        feature_shape = global_feature_h5f[feature_name].shape
        if len(feature_shape) == 1:
            feature_shape += (1,)
        feature_shapes.append(feature_shape)
    max_shape_length = max(map(len, feature_shapes))
    if max_shape_length > 2:
        raise NotImplementedError("tensor feature is not supported yet")

    n_instances = feature_shapes[0][0]
    for feature_shape in feature_shapes:
        if feature_shape[0] != n_instances:
            raise ValueError("different number of instances: {} and {}."
                             .format(feature_shapes[0], feature_shape))
    n_features = sum(shape[1] for shape in feature_shapes)
    concat_shape = (n_instances, n_features)

    concat_feature_h5f.create_dataset('feature', shape=concat_shape,
                                      dtype=np.float32)

    feature_d = 0
    dset = concat_feature_h5f['feature']
    for feature_i, (feature_name, feature_shape) in enumerate(
            zip(feature_list, feature_shapes)):
        batch_size = (buffer_size
                      // (global_feature_h5f[feature_name].dtype.itemsize
                          * feature_shape[1]))
        if batch_size == 0:
            print("Warning! buffer_size not enough to fitted by an "
                  "instance. Trying to use more memory.")
            batch_size = 1
        with IterTimer("({}/{}) Filling {}"
                       .format(feature_i + 1, len(feature_list), feature_name),
                       feature_shape[0]) as timer:
            for batch_start in range(0, feature_shape[0], batch_size):
                timer.update(batch_start)
                batch_end = min(feature_shape[0], batch_start + batch_size)
                feature_buffer = (global_feature_h5f[feature_name]
                                  [batch_start: batch_end])
                if len(feature_buffer.shape) == 1:
                    feature_buffer = feature_buffer[:, np.newaxis]
                dset[batch_start: batch_end,
                     feature_d: feature_d + feature_shape[1]] = feature_buffer

        feature_d += feature_shape[1]


def save_concat_features(feature_config, global_feature_hdf_path,
                         buffer_size=int(1e+9)):

    concat_feature_hdf_path = feature_config.concat_feature_hdf_path
    feature_list = feature_config.feature_list

    with h5py.File(global_feature_hdf_path, 'r') as global_feature_h5f, \
            h5py.File(concat_feature_hdf_path, 'w') as concat_feature_h5f, \
            SimpleTimer("Concating and saving features", end_in_new_line=False):
        # save label
        for label in feature_config.label_list:
            concat_feature_h5f.create_dataset(
                'label/' + label, data=global_feature_h5f[label])

        # save extra data
        for data_name in feature_config.extra_data_list:
            concat_feature_h5f.create_dataset(
                'extra/' + data_name, data=global_feature_h5f[data_name])

        # save extra features
        fill_concat_features(feature_list, global_feature_h5f,
                             concat_feature_h5f, buffer_size)

    gc.collect()



from __future__ import print_function
import gc
from functools import wraps, partial

import numpy as np
from bistiming import SimpleTimer


def can_skip(set_name, new_data_name, interm_data, global_feature_h5f,
             show_skip):
    if set_name == "intermediate_data":
        key = new_data_name
        generated_set = interm_data
    elif set_name == "features":
        key = "/" + new_data_name
        generated_set = global_feature_h5f.keys()
    else:
        raise NotImplementedError()

    if key in generated_set:
        if show_skip:
            print("skip {} {}: already generated".format(
                set_name, new_data_name))
        return True
    return False


def data_type(set_name, func_dict, skip_if_exist=True, show_skip=True):
    def intermediate_data_decorator(args):
        func, will_generate_keys = args
        if func.__name__ in func_dict.values():
            raise ValueError("duplicated function name " + func.__name__)
        for key in will_generate_keys:
            if key in func_dict:
                raise ValueError("duplicated {} {} in {} and {}".format(
                    set_name, key, func_dict[key], func.__name__))
            func_dict[key] = func.__name__

        @wraps(func)
        def func_wrapper(self, new_data_name):
            if skip_if_exist and can_skip(
                    set_name, new_data_name, self.intermediate_data,
                    self.global_feature_h5f, show_skip):
                return
            func(self, set_name, new_data_name)
            gc.collect()
        return func_wrapper
    return intermediate_data_decorator


def require_intermediate_data(data_names, interm_data_func_dict):
    if isinstance(data_names, str):
        data_names = (data_names,)

    def require_intermediate_data_decorator(args):
        func, will_generate_keys = args

        @wraps(func)
        def func_wrapper(self, set_name, new_data_name):
            self.require(data_names, interm_data_func_dict)
            data = {key: self.intermediate_data[key] for key in data_names}
            func(self, set_name, new_data_name, data=data)
        return func_wrapper, will_generate_keys
    return require_intermediate_data_decorator


def will_generate(will_generate_keys, manually_create_dataset=False):
    if isinstance(will_generate_keys, str):
        will_generate_keys = (will_generate_keys,)
    will_generate_key_set = set(will_generate_keys)

    def will_generate_decorator(func):
        @wraps(func)
        def func_wrapper(self, set_name, new_data_name, **kwargs):
            if manually_create_dataset:
                kwargs['create_dataset_functions'] = {
                    k: partial(self.global_feature_h5f.create_dataset, k)
                    for k in will_generate_key_set
                }
            with SimpleTimer("Generating {} {} using {}"
                             .format(set_name, new_data_name, func.__name__),
                             end_in_new_line=False):  # pylint: disable=C0330
                result_dict = func(self, **kwargs)
            if result_dict is None:
                result_dict = {}
            if not (hasattr(result_dict, 'keys')
                    and hasattr(result_dict, '__getitem__')):
                raise ValueError("the return value of mehod {} should have "
                                 "keys and __getitem__ methods".format(
                                     func.__name__))

            result_dict_key_set = set(result_dict.keys())
            if manually_create_dataset:
                redundant_key_set = result_dict_key_set - will_generate_key_set
                if len(redundant_key_set) > 0:
                    raise ValueError("the return keys of method {} {} have "
                                     "redundant keys {} while generating {} {}"
                                     .format(
                                         func.__name__, result_dict_key_set,
                                         redundant_key_set, set_name,
                                         new_data_name))
                # TODO: check all the datasets is either manually created or in
                #       result_dict_key_set
            elif will_generate_key_set != result_dict_key_set:
                raise ValueError("the return keys of method {} {} doesn't "
                                 "match {} while generating {} {}".format(
                                     func.__name__, result_dict_key_set,
                                     will_generate_key_set, set_name,
                                     new_data_name))

            if set_name == "intermediate_data":
                self.intermediate_data.update(result_dict)
            elif set_name == "features":
                for key in result_dict_key_set:
                    if np.isnan(result_dict[key]).any():
                        raise ValueError("features {} have nan".format(key))
                    with SimpleTimer("Writing generated features {} "
                                     "to hdf5 file".format(key),
                                     end_in_new_line=False):
                        if key in self.global_feature_h5f:
                            self.global_feature_h5f[key][...] = result_dict[key]
                        else:
                            self.global_feature_h5f.create_dataset(
                                key, data=result_dict[key])
            else:
                raise NotImplementedError()
        return func_wrapper, will_generate_keys
    return will_generate_decorator


intermediate_data = partial(data_type, "intermediate_data")
features = partial(data_type, "features")



import os.path

import h5py
from mkdir_p import mkdir_p


class FeatureGenerator(object):

    def __init__(self, global_feature_hdf_path, feature_func_dict):
        self.intermediate_data = {}
        mkdir_p(os.path.dirname(global_feature_hdf_path))
        self.global_feature_h5f = h5py.File(global_feature_hdf_path, 'a')
        self.feature_func_dict = feature_func_dict

    def generate(self, feature_names):
        if isinstance(feature_names, str):
            feature_names = [feature_names]
        self.require(feature_names, self.feature_func_dict)

    def require(self, keys, func_dict):
        for key in keys:
            getattr(self, func_dict[key])(key)



import pkg_resources
from .feature_generator import FeatureGenerator
from .concat_features import save_concat_features

__all__ = ['feature_generator', 'decorators', 'concat_features']
__version__ = pkg_resources.get_distribution("feagen").version



#!/usr/bin/env python
import os
from setuptools import setup, find_packages


on_rtd = os.environ.get('READTHEDOCS', None) == 'True'
# read the docs could not compile numpy and c extensions
if on_rtd:
    setup_requires = []
    install_requires = []
    tests_require = []
else:
    setup_requires = [
        'nose',
        'coverage',
    ]
    install_requires = [
        'mkdir-p',
        'h5py',
        'bistiming',
        'numpy',
    ]
    tests_require = []

description = ("A fast and memory-efficient Python feature generating "
               "framework for machine learning.")
long_description = ("See `github <https://github.com/ianlini/feagen>`_ "
                    "for more information.")

setup(
    name='feagen',
    version="0.1.0",
    description=description,
    long_description=long_description,
    author='ianlini',
    url='https://github.com/ianlini/feagen',
    setup_requires=setup_requires,
    install_requires=install_requires,
    tests_require=tests_require,
    classifiers=[
        'Topic :: Utilities',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
    ],
    test_suite='nose.collector',
    packages=find_packages(),
)
